(*
///////////////////////////////////////////////////////////////////////////////
// glInfo.ml
// =========
// get GL vendor, version, supported extensions and other states using glGet*
// functions and store them glInfo struct variable
//
// To get valid OpenGL infos, OpenGL rendering context (RC) must be opened
// before calling glInfo::getInfo(). Otherwise it returns false.
//
//  AUTHOR: Song Ho Ahn (song.ahn@gmail.com)
// CREATED: 2005-10-04
// UPDATED: 2009-10-06
//
// Copyright (c) 2005 Song Ho Ahn
///////////////////////////////////////////////////////////////////////////////
*)

open GL

let vendor   = ref ""
let renderer = ref ""
let version  = ref ""
let extensions = ref []

let redBits   = ref 0
let greenBits = ref 0
let blueBits  = ref 0
let alphaBits = ref 0

let depthBits = ref 0
let stencilBits = ref 0
let maxLights = ref 0
let maxTextureSize = ref 0
let maxClipPlanes = ref 0
let maxModelViewStacks = ref 0
let maxProjectionStacks = ref 0
let maxAttribStacks = ref 0
let maxTextureStacks = ref 0

(*
///////////////////////////////////////////////////////////////////////////////
// extract openGL info
// This function must be called after GL rendering context opened.
///////////////////////////////////////////////////////////////////////////////
*)
let getInfo() =

  (* get vendor string *)
  vendor := glGetString GL_VENDOR;

  (* get renderer string *)
  renderer := glGetString GL_RENDERER;

  (* get version string *)
  version := glGetString GL_VERSION;

  (* get all extensions as a string *)
  let str_exts = glGetString GL_EXTENSIONS in
  (* split extensions *)
  let ext_li = Str.split (Str.regexp " ") str_exts in

  (* sort extension by alphabetical order *)
  extensions := List.sort compare ext_li;

  (* get number of color bits *)
  redBits   := glGetInteger1 Get.GL_RED_BITS;
  greenBits := glGetInteger1 Get.GL_GREEN_BITS;
  blueBits  := glGetInteger1 Get.GL_BLUE_BITS;
  alphaBits := glGetInteger1 Get.GL_ALPHA_BITS;

  (* get depth bits *)
  depthBits := glGetInteger1 Get.GL_DEPTH_BITS;

  (* get stecil bits *)
  stencilBits := glGetInteger1 Get.GL_STENCIL_BITS;

  (* get max number of lights allowed *)
  maxLights := glGetInteger1 Get.GL_MAX_LIGHTS;

  (* get max texture resolution *)
  maxTextureSize := glGetInteger1 Get.GL_MAX_TEXTURE_SIZE;

  (* get max number of clipping planes *)
  maxClipPlanes := glGetInteger1 Get.GL_MAX_CLIP_PLANES;

  (* get max modelview and projection matrix stacks *)
  maxModelViewStacks := glGetInteger1 Get.GL_MAX_MODELVIEW_STACK_DEPTH;
  maxProjectionStacks := glGetInteger1 Get.GL_MAX_PROJECTION_STACK_DEPTH;
  maxAttribStacks := glGetInteger1 Get.GL_MAX_ATTRIB_STACK_DEPTH;

  (* get max texture stacks *)
  maxTextureStacks := glGetInteger1 Get.GL_MAX_TEXTURE_STACK_DEPTH;
;;


(*
///////////////////////////////////////////////////////////////////////////////
// check if the video card support a certain extension
///////////////////////////////////////////////////////////////////////////////
*)
let isExtensionSupported ext =
  (* search corresponding extension *)
  (List.mem ext !extensions)
;;


(*
///////////////////////////////////////////////////////////////////////////////
// print OpenGL info to screen and save to a file
///////////////////////////////////////////////////////////////////////////////
*)
let printSelf() =

  Printf.printf "\n";  (* blank line *)
  Printf.printf "OpenGL Driver Info\n";
  Printf.printf "==================\n";
  Printf.printf "Vendor: %s\n" !vendor;
  Printf.printf "Version: %s\n" !version;
  Printf.printf "Renderer: %s\n" !renderer;

  Printf.printf "\n";
  Printf.printf "Color Bits(R,G,B,A): (%d, %d, %d, %d)\n" !redBits !greenBits !blueBits !alphaBits;
  Printf.printf "Depth Bits: %d\n" !depthBits;
  Printf.printf "Stencil Bits: %d\n" !stencilBits;

  Printf.printf "\n";
  Printf.printf "Max Texture Size: %dx%d\n" !maxTextureSize !maxTextureSize;
  Printf.printf "Max Lights: %d\n" !maxLights;
  Printf.printf "Max Clip Planes: %d\n" !maxClipPlanes;
  Printf.printf "Max Modelview Matrix Stacks: %d\n" !maxModelViewStacks;
  Printf.printf "Max Projection Matrix Stacks: %d\n" !maxProjectionStacks;
  Printf.printf "Max Attribute Stacks: %d\n" !maxAttribStacks;
  Printf.printf "Max Texture Stacks: %d\n" !maxTextureStacks;

  Printf.printf "\n";
  Printf.printf "Total Number of Extensions: %d\n" (List.length !extensions);
  Printf.printf "==============================\n";
  List.iter (Printf.printf " %s\n") !extensions;
  Printf.printf "======================================================================\n";
  Printf.printf "\n";
;;

