(*
///////////////////////////////////////////////////////////////////////////////
// main.cpp
// ========
// testing buffer object for vertex, GL_ARB_vertex_buffer_object extension
// draw a cube using glVertexArrays()
//
//  AUTHOR: Song Ho Ahn (song.ahn@gmail.com)
// CREATED: 2006-11-14
// UPDATED: 2006-11-14
///////////////////////////////////////////////////////////////////////////////
*)

open GL
open Glu
open Glut
open VertArray
open VBO

(*
///////////////////////////////////////////////////////////////////////////////
// initialize global variables
///////////////////////////////////////////////////////////////////////////////
*)
let font = GLUT_BITMAP_8_BY_13
let vboId = ref None                   (* ID of VBO for vertex arrays *)
let mouseLeftDown = ref false
let mouseRightDown = ref false
let mouseX = ref 0
let mouseY = ref 0
let cameraAngleX = ref 0.0
let cameraAngleY = ref 0.0
let cameraDistance = ref 0.0
let vboSupported = ref false
let vboUsed = ref false
let drawMode = ref 0

let ( += ) a b = (a := !a +. b)

let ( !! ) a = match !a with Some a -> a | None -> invalid_arg "uninitialised var" ;;

(*
// cube ///////////////////////////////////////////////////////////////////////
//    v6----- v5
//   /|      /|
//  v1------v0|
//  | |     | |
//  | |v7---|-|v4
//  |/      |/
//  v2------v3
*)

(* vertex coords array *)
let vertices = [| 1; 1; 1;  -1; 1; 1;  -1;-1; 1;   1;-1; 1;    (* v0-v1-v2-v3 *)
                  1; 1; 1;   1;-1; 1;   1;-1;-1;   1; 1;-1;    (* v0-v3-v4-v5 *)
                  1; 1; 1;   1; 1;-1;  -1; 1;-1;  -1; 1; 1;    (* v0-v5-v6-v1 *)
                 -1; 1; 1;  -1; 1;-1;  -1;-1;-1;  -1;-1; 1;    (* v1-v6-v7-v2 *)
                 -1;-1;-1;   1;-1;-1;   1;-1; 1;  -1;-1; 1;    (* v7-v4-v3-v2 *)
                  1;-1;-1;  -1;-1;-1;  -1; 1;-1;   1; 1;-1; |] (* v4-v7-v6-v5 *)

(* normal array *)
let normals = [|  0; 0; 1;   0; 0; 1;   0; 0; 1;   0; 0; 1;    (* v0-v1-v2-v3 *)
                  1; 0; 0;   1; 0; 0;   1; 0; 0;   1; 0; 0;    (* v0-v3-v4-v5 *)
                  0; 1; 0;   0; 1; 0;   0; 1; 0;   0; 1; 0;    (* v0-v5-v6-v1 *)
                 -1; 0; 0;  -1; 0; 0;  -1; 0; 0;  -1; 0; 0;    (* v1-v6-v7-v2 *)
                  0;-1; 0;   0;-1; 0;   0;-1; 0;   0;-1; 0;    (* v7-v4-v3-v2 *)
                  0; 0;-1;   0; 0;-1;   0; 0;-1;   0; 0;-1; |] (* v4-v7-v6-v5 *)

(* color array *)
let colors = [| 1;1;1;  1;1;0;  1;0;0;  1;0;1;                 (* v0-v1-v2-v3 *)
                1;1;1;  1;0;1;  0;0;1;  0;1;1;                 (* v0-v3-v4-v5 *)
                1;1;1;  0;1;1;  0;1;0;  1;1;0;                 (* v0-v5-v6-v1 *)
                1;1;0;  0;1;0;  0;0;0;  1;0;0;                 (* v1-v6-v7-v2 *)
                0;0;0;  0;0;1;  1;0;1;  1;0;0;                 (* v7-v4-v3-v2 *)
                0;0;1;  0;0;0;  0;1;0;  0;1;1; |]              (* v4-v7-v6-v5 *)

(* convert to bigarray buffers which are binary compatible with C *)
let to_ba = Bigarray.Array1.of_array Bigarray.float32 Bigarray.c_layout ;;

let vertices = to_ba (Array.map float_of_int vertices)
let normals = to_ba (Array.map float_of_int normals)
let colors = to_ba (Array.map float_of_int colors)

(*
///////////////////////////////////////////////////////////////////////////////
// write 2d text using GLUT
// The projection matrix must be set to orthogonal before call this function.
///////////////////////////////////////////////////////////////////////////////
*)
let drawString str x y color font =
  glPushAttrib [Attrib.GL_LIGHTING_BIT; Attrib.GL_CURRENT_BIT]; (* lighting and color mask *)
  glDisable GL_LIGHTING;      (* need to disable lighting for proper text color *)

  glColor4v color;            (* set text color *)
  glRasterPos2i ~x ~y;        (* place text position *)

  (* loop all characters in the string *)
  String.iter (fun c ->
    glutBitmapCharacter ~font ~c;
  ) str;

  glEnable GL_LIGHTING;
  glPopAttrib();
;;


(*
///////////////////////////////////////////////////////////////////////////////
// draw a string in 3D space
///////////////////////////////////////////////////////////////////////////////
*)
let drawString3D str (x,y,z) color font =
  glPushAttrib [Attrib.GL_LIGHTING_BIT; Attrib.GL_CURRENT_BIT];  (* lighting and color mask *)
  glDisable GL_LIGHTING;      (* need to disable lighting for proper text color *)

  glColor4v color;            (* set text color *)
  glRasterPos3 ~x ~y ~z;      (* place text position *)

  (* loop all characters in the string *)
  String.iter (fun c ->
    glutBitmapCharacter ~font ~c;
  ) str;

  glEnable GL_LIGHTING;
  glPopAttrib();
;;


(*
///////////////////////////////////////////////////////////////////////////////
// destroy a VBO
// If VBO id is not valid or zero, then OpenGL ignores it silently.
///////////////////////////////////////////////////////////////////////////////
*)
let deleteVBO vboId =
  glDeleteBuffer vboId;
;;


(*
///////////////////////////////////////////////////////////////////////////////
// clean up shared memory
///////////////////////////////////////////////////////////////////////////////
*)
let clearSharedMem() =
  (* clean up VBOs *)
  if !vboSupported then
    deleteVBO !!vboId;
;;


(*
///////////////////////////////////////////////////////////////////////////////
// set camera position and lookat direction
///////////////////////////////////////////////////////////////////////////////
*)
let setCamera posX posY posZ targetX targetY targetZ =
  glMatrixMode GL_MODELVIEW;
  glLoadIdentity();
  gluLookAt posX posY posZ  targetX targetY targetZ  0. 1. 0.; (* eye(x,y,z), focal(x,y,z), up(x,y,z) *)
;;


(*
///////////////////////////////////////////////////////////////////////////////
// generate vertex buffer object and bind it with its data
// You must give 2 hints about data usage; target and mode, so that OpenGL can
// decide which data should be stored and its location.
// VBO works with 2 different targets; GL_ARRAY_BUFFER for vertex arrays
// and GL_ELEMENT_ARRAY_BUFFER for index array in glDrawElements().
// The default target is GL_ARRAY_BUFFER.
// By default, usage mode is set as GL_STATIC_DRAW.
// Other usages are GL_STREAM_DRAW, GL_STREAM_READ, GL_STREAM_COPY,
// GL_STATIC_DRAW, GL_STATIC_READ, GL_STATIC_COPY, GL_DYNAMIC_DRAW,
// GL_DYNAMIC_READ, GL_DYNAMIC_COPY.
///////////////////////////////////////////////////////////////////////////////
*)
let createVBO data dataSize target usage =
  (* if glGenBuffers() fails to create a VBO-id, it raises (Failure "glGenBuffer") *)
  let id = glGenBuffer() in                 (* create a vbo *)
  glBindBuffer target id;                   (* activate vbo id to use *)
  glBufferData target dataSize data usage;  (* upload data to video card *)

  (* check data size in VBO is same as input array, if not return 0 and delete VBO *)
  let bufferSize = glGetBufferSize target in
  if(dataSize = bufferSize)
  then (id)      (* return VBO id *)
  else begin
    glDeleteBuffer id;
    failwith "[createVBO()] Data size is mismatch with input array";
  end;
;;


(*
///////////////////////////////////////////////////////////////////////////////
// display info messages
///////////////////////////////////////////////////////////////////////////////
*)
let showInfo() =
  (* backup current model-view matrix *)
  glPushMatrix();                  (* save current modelview matrix *)
  glLoadIdentity();                (* reset modelview matrix *)

  (* set to 2D orthogonal projection *)
  glMatrixMode GL_PROJECTION;      (* switch to projection matrix *)
  glPushMatrix();                  (* save current projection matrix *)
  glLoadIdentity();                (* reset projection matrix *)
  gluOrtho2D 0. 400. 0. 300.;      (* set to orthogonal projection *)

  let color = (1., 1., 1., 1.) in

  let ss = "VBO: " ^ (if !vboUsed then "on" else "off") in  (* add 0(ends) at the end *)
  drawString ss 1 286 color font;

  let ss = "Press SPACE key to toggle VBO on/off." in
  drawString ss 1 1 color font;

  (* restore projection matrix *)
  glPopMatrix();                   (* restore to previous projection matrix *)

  (* restore modelview matrix *)
  glMatrixMode GL_MODELVIEW;       (* switch to modelview matrix *)
  glPopMatrix();                   (* restore to previous modelview matrix *)
;;



(*
//=============================================================================
// CALLBACKS
//=============================================================================
*)
let displayCB() =
  (* clear buffer *)
  glClear [GL_COLOR_BUFFER_BIT; GL_DEPTH_BUFFER_BIT; GL_STENCIL_BUFFER_BIT];

  (* save the initial ModelView matrix before modifying ModelView matrix *)
  glPushMatrix();

  (* tramsform camera *)
  glTranslate 0. 0. !cameraDistance;
  glRotate !cameraAngleX 1. 0. 0.;   (* pitch *)
  glRotate !cameraAngleY 0. 1. 0.;   (* heading *)

  if !vboUsed then  (* draw cube using VBO *)
  begin
    (* bind VBOs with IDs and set the buffer offsets of the bound VBOs *)
    (* When buffer object is bound with its ID, all pointers in gl*Pointer() *)
    (* are treated as offset instead of real pointer. *)
    glBindBuffer GL_ARRAY_BUFFER !!vboId;

    (* enable vertex arrays *)
    glEnableClientState GL_NORMAL_ARRAY;
    glEnableClientState GL_COLOR_ARRAY;
    glEnableClientState GL_VERTEX_ARRAY;

    (* before draw, specify vertex and index arrays with their offsets *)
    glNormalPointerOfs8 Norm.GL_FLOAT 0 (ba_sizeof vertices);
    glColorPointerOfs8 3 Color.GL_FLOAT 0 ((ba_sizeof vertices) + (ba_sizeof normals));
    glVertexPointerOfs8 3 Coord.GL_FLOAT 0 0;

    glDrawArrays GL_QUADS 0 24;

    glDisableClientState GL_VERTEX_ARRAY;  (* disable vertex arrays *)
    glDisableClientState GL_COLOR_ARRAY;
    glDisableClientState GL_NORMAL_ARRAY;

    (* it is good idea to release VBOs with ID 0 after use. *)
    (* Once bound with 0, all pointers in gl*Pointer() behave as real *)
    (* pointer, so, normal vertex array operations are re-activated *)
    glUnbindBuffer GL_ARRAY_BUFFER;
  end

  else        (* draw a cube using vertex array method *)
  (* notice that only difference between VBO and VA is binding buffers and offsets *)
  begin
    (* enable vertex arrays *)
    glEnableClientState GL_NORMAL_ARRAY;
    glEnableClientState GL_COLOR_ARRAY;
    glEnableClientState GL_VERTEX_ARRAY;

    (* before draw, specify vertex arrays *)
    glNormalPointer Norm.GL_FLOAT 0 normals;
    glColorPointer 3 Color.GL_FLOAT 0 colors;
    glVertexPointer 3 Coord.GL_FLOAT 0 vertices;

    glDrawArrays GL_QUADS 0 24;

    glDisableClientState GL_VERTEX_ARRAY;  (* disable vertex arrays *)
    glDisableClientState GL_COLOR_ARRAY;
    glDisableClientState GL_NORMAL_ARRAY;
  end;

  (* draw info messages *)
  showInfo();

  glPopMatrix();

  glutSwapBuffers();
;;


let reshapeCB ~width:w ~height:h =
  (* set viewport to be the entire window *)
  glViewport 0 0 w h;

  (* set perspective viewing frustum *)
  let aspectRatio = (float w /. float h) in
  glMatrixMode GL_PROJECTION;
  glLoadIdentity();
  (*glFrustum (-. aspectRatio) aspectRatio (-1.) 1. 1. 100.; *)
  gluPerspective 60.0 aspectRatio 1.0 1000.0;  (* FOV, AspectRatio, NearClip, FarClip *)

  (* switch to modelview matrix in order to set scene *)
  glMatrixMode GL_MODELVIEW;
;;


let rec timerCB ~value:millisec =
  glutTimerFunc millisec timerCB millisec;
  glutPostRedisplay();
;;


let idleCB() =
  glutPostRedisplay();
;;


let keyboardCB ~key ~x ~y =
  begin match key with
  | '\027' ->  (* ESCAPE *)
      exit(0);

  | ' ' ->
      if !vboSupported then
        vboUsed := not !vboUsed;

  | 'd'     (* switch rendering modes (fill -> wire -> point) *)
  | 'D' ->
      drawMode := (!drawMode + 1) mod 3;
      if !drawMode = 0 then       (* fill mode *)
      begin
        glPolygonMode GL_FRONT_AND_BACK GL_FILL;
        glEnable GL_DEPTH_TEST;
        glEnable GL_CULL_FACE;
      end
      else if !drawMode = 1 then  (* wireframe mode *)
      begin
        glPolygonMode GL_FRONT_AND_BACK GL_LINE;
        glDisable GL_DEPTH_TEST;
        glDisable GL_CULL_FACE;
      end
      else                    (* point mode *)
      begin
        glPolygonMode GL_FRONT_AND_BACK GL_POINT;
        glDisable GL_DEPTH_TEST;
        glDisable GL_CULL_FACE;
      end
  | _ -> ()
  end;

  glutPostRedisplay();
;;



let mouseCB ~button ~state ~x ~y =
  mouseX := x;
  mouseY := y;

  match button, state with
  | GLUT_LEFT_BUTTON, GLUT_DOWN -> mouseLeftDown := true
  | GLUT_LEFT_BUTTON, GLUT_UP -> mouseLeftDown := false

  | GLUT_RIGHT_BUTTON, GLUT_DOWN -> mouseRightDown := true
  | GLUT_RIGHT_BUTTON, GLUT_UP -> mouseRightDown := false
  | _ -> ()
;;


let mouseMotionCB ~x ~y =
  if !mouseLeftDown then
  begin
    cameraAngleY += float (x - !mouseX);
    cameraAngleX += float (y - !mouseY);
    mouseX := x;
    mouseY := y;
  end;

  if !mouseRightDown then
  begin
    cameraDistance += float (y - !mouseY) *. 0.2;
    mouseY := y;
  end;

  (*glutPostRedisplay(); *)
;;



(*
///////////////////////////////////////////////////////////////////////////////
// initialize lights
///////////////////////////////////////////////////////////////////////////////
*)
let initLights() =
  (* set up light colors (ambient, diffuse, specular) *)
  let lightKa = (0.2, 0.2, 0.2, 1.0)  (* ambient light *)
  and lightKd = (0.7, 0.7, 0.7, 1.0)  (* diffuse light *)
  and lightKs = (1.0, 1.0, 1.0, 1.0)  (* specular light *)
  in
  glLight (GL_LIGHT 0) (Light.GL_AMBIENT lightKa);
  glLight (GL_LIGHT 0) (Light.GL_DIFFUSE lightKd);
  glLight (GL_LIGHT 0) (Light.GL_SPECULAR lightKs);

  (* position the light *)
  let lightPos = (0., 0., 20., 1.) in  (* positional light *)
  glLight (GL_LIGHT 0) (Light.GL_POSITION lightPos);

  glEnable GL_LIGHT0;  (* MUST enable each light source after configuration *)
;;


(*
///////////////////////////////////////////////////////////////////////////////
// initialize GLUT for windowing
///////////////////////////////////////////////////////////////////////////////
*)
let initGLUT() =
  (* GLUT stuff for windowing *)
  (* initialization openGL window. *)
  (* it is called before any other GLUT routine *)
  ignore(glutInit Sys.argv);

  glutInitDisplayMode [GLUT_RGB; GLUT_DOUBLE; GLUT_DEPTH; GLUT_STENCIL];   (* display mode *)

  glutInitWindowSize 400 300;               (* window size *)
  glutInitWindowPosition 100 100;           (* window location *)

  (* finally, create a window with openGL context *)
  (* Window will not displayed until glutMainLoop() is called *)
  (* it returns a unique ID *)
  let handle = glutCreateWindow Sys.argv.(0) in   (* param is the title of window *)

  (* register GLUT callback functions *)
  glutDisplayFunc displayCB;
  glutTimerFunc 33 timerCB 33;             (* redraw only every given millisec *)
  (*glutIdleFunc(idleCB); *)               (* redraw when idle *)
  glutReshapeFunc reshapeCB;
  glutKeyboardFunc keyboardCB;
  glutMouseFunc mouseCB;
  glutMotionFunc mouseMotionCB;

  (handle)
;;


(*
///////////////////////////////////////////////////////////////////////////////
// initialize OpenGL
// disable unused features
///////////////////////////////////////////////////////////////////////////////
*)
let initGL() =
  glShadeModel GL_SMOOTH;                    (* shading mathod: GL_SMOOTH or GL_FLAT *)
  glPixelStorei GL_UNPACK_ALIGNMENT 4;       (* 4-byte pixel alignment *)

  (* enable /disable features *)
  glHint GL_PERSPECTIVE_CORRECTION_HINT  GL_NICEST;
  (*glHint GL_LINE_SMOOTH_HINT  GL_NICEST; *)
  (*glHint GL_POLYGON_SMOOTH_HINT  GL_NICEST; *)
  glEnable GL_DEPTH_TEST;
  glEnable GL_LIGHTING;
  glEnable GL_TEXTURE_2D;
  glEnable GL_CULL_FACE;

   (* track material ambient and diffuse from surface color, call it before glEnable(GL_COLOR_MATERIAL) *)
  glColorMaterial GL_FRONT_AND_BACK  GL_AMBIENT_AND_DIFFUSE;
  glEnable GL_COLOR_MATERIAL;

  glClearColor 0. 0. 0. 0.;                  (* background color *)
  glClearStencil 0;                          (* clear stencil buffer *)
  glClearDepth 1.0;                          (* 0 is near, 1 is far *)
  glDepthFunc GL_LEQUAL;

  initLights();
  setCamera 0. 0. 5. 0. 0. 0.;
;;


let exitCB() =
  clearSharedMem();
;;


(* main /////////////////////////////////////////////////////////////////////// *)
let () =
  (* init GLUT and GL *)
  let _ = initGLUT() in
  initGL();

  (* register exit callback *)
  at_exit exitCB;

  (* get OpenGL info *)
  GlInfo.getInfo();
  (*GlInfo.printSelf(); *)

  if GlInfo.isExtensionSupported "GL_ARB_vertex_buffer_object" then
  begin
    vboSupported := true;
    vboUsed := true;
    print_endline "Video card supports GL_ARB_vertex_buffer_object.";
  end else begin
    vboSupported := false;
    vboUsed := false;
    prerr_endline "Video card does NOT support GL_ARB_vertex_buffer_object.";
  end;


  if !vboSupported then
  begin
    (* create vertex buffer objects, you need to delete them when program exits *)
    (* Try to put both vertex coords array, vertex normal array and vertex color in the same buffer object. *)
    (* glBufferDataARB with NULL pointer reserves only memory space. *)
    (* Copy actual data with 2 calls of glBufferSubDataARB, one for vertex coords and one for normals. *)
    (* target flag is GL_ARRAY_BUFFER, and usage flag is GL_STATIC_DRAW_ARB *)
    vboId := Some(glGenBuffer());
    glBindBuffer GL_ARRAY_BUFFER !!vboId;
    let vertices_size = ba_sizeof vertices
    and normals_size  = ba_sizeof normals
    and colors_size   = ba_sizeof colors
    in
    glBufferDataNull GL_ARRAY_BUFFER (vertices_size + normals_size + colors_size) GL_STATIC_DRAW;
    glBufferSubData GL_ARRAY_BUFFER 0 (vertices_size) vertices;                        (* copy vertices starting from 0 offest *)
    glBufferSubData GL_ARRAY_BUFFER (vertices_size) (normals_size) normals;            (* copy normals after vertices *)
    glBufferSubData GL_ARRAY_BUFFER (vertices_size + normals_size) colors_size colors; (* copy colours after normals *)
  end;

  (* the last GLUT call (LOOP) *)
  (* window will be shown and display callback is triggered by events *)
  (* NOTE: this call never return main(). *)
  glutMainLoop();  (* Start GLUT event-processing loop *)
;;

